from dataclasses import dataclass

@dataclass
class LoraSettings:
    """Basic LoRA training settings used by all training regimes"""
    rank: int
    alpha: int
    num_epochs: int
    dropout: float = 0.01
    weight_decay: float = 0.01
    learning_rate: float = 4e-4
    evals_per_epoch: int = 1
    wandb_project: str | None = None
    wandb_api_key: str | None = None



from typing import Sequence, TypedDict, Literal
from dataclasses import dataclass, field, asdict, replace
import yaml
import os
from .datasets import Dataset, Convos, HubMessageConvos, JsonlConvos, HubInstructConvos
from .lora import LoraSettings

@dataclass
class LoraCloudTrainer:
    provider: Literal["modal"]
    timeout: int

def llama_3_1_70b_axolotl(
    dataset: Dataset[Convos],
    settings: LoraSettings,
    cloud: LoraCloudTrainer | None = None,
    warmup_steps=10,
):
    """Given a dataset and some settings, sets up an Axolotl config for training
    Llama 3.1 70B Instruct"""
    config = Config(
        base_model="unsloth/Meta-Llama-3.1-70B-Instruct",
        model_type="LlamaForCausalLM",
        lora_r=settings.rank,
        lora_alpha=settings.alpha,
        lora_dropout=settings.dropout,
        gradient_accumulation_steps=16,
        micro_batch_size=1,
        num_epochs=settings.num_epochs,
        learning_rate=settings.learning_rate,
        warmup_steps=warmup_steps,
        evals_per_epoch=settings.evals_per_epoch,
        weight_decay=settings.weight_decay,
        fsdp_config=generate_fsdp_conf("LlamaDecoderLayer"),
        wandb_project=settings.wandb_project,
    )
    return TrainingConfig(
        axolotl_config=config.generate(dataset),
        cloud_config=None if not cloud else CloudConfig(
            provider=cloud.provider,
            gpu="h100",
            gpu_count=8,
            timeout=cloud.timeout,
        ),
    )

def llama_3_1_8b_axolotl(
    dataset: Dataset[Convos],
    settings: LoraSettings,
    cloud: LoraCloudTrainer | None = None,
    warmup_steps: int = 10,
):
    """Given a dataset and some settings, sets up an Axolotl config for training
    Llama 3.1 8B Instruct"""
    config = Config(
        base_model="unsloth/Meta-Llama-3.1-8B-Instruct",
        model_type="LlamaForCausalLM",
        lora_r=settings.rank,
        lora_alpha=settings.alpha,
        lora_dropout=settings.dropout,
        gradient_accumulation_steps=4,
        micro_batch_size=4,
        num_epochs=settings.num_epochs,
        learning_rate=settings.learning_rate,
        warmup_steps=warmup_steps,
        evals_per_epoch=settings.evals_per_epoch,
        weight_decay=settings.weight_decay,
        fsdp_config=None,
        wandb_project=settings.wandb_project,
    )
    return TrainingConfig(
        axolotl_config=config.generate(dataset),
        cloud_config=None if not cloud else CloudConfig(
            provider=cloud.provider,
            gpu="h100",
            gpu_count=1,
            timeout=cloud.timeout,
        ),
    )

@dataclass
class AxolotlJsonlConvos:
    path: str
    ds_type: str = "json"
    split: str = "train"
    type: str = "chat_template"

class PropertyMappings(TypedDict):
    role: str
    content: str

class RoleMapping(TypedDict):
    user: list[str]
    assistant: list[str]
    system: list[str]

@dataclass
class AxolotlHubConvos:
    path: str
    field_messages: str
    message_property_mappings: PropertyMappings
    roles: RoleMapping
    type: str = "chat_template"
    chat_template: str = "tokenizer_default"

class InstructType(TypedDict):
    field_instruction: str
    field_input: str
    field_output: str

@dataclass
class AxolotlHubInstructions:
    path: str
    type: InstructType

class FsdpConf(TypedDict):
    fsdp_limit_all_gathers: bool
    fsdp_sync_module_states: bool
    fsdp_offload_params: bool
    fsdp_use_orig_params: bool
    fsdp_cpu_ram_efficient_loading: bool
    fsdp_auto_wrap_policy: str
    fsdp_transformer_layer_cls_to_wrap: str
    fsdp_state_dict_type: str
    fsdp_sharding_strategy: str

def generate_fsdp_conf(layer_to_wrap: str) -> FsdpConf:
    """Generates an FSDP config based on the layer to wrap"""
    return {
        "fsdp_limit_all_gathers": True,
        "fsdp_sync_module_states": True,
        "fsdp_offload_params": True,
        "fsdp_use_orig_params": False,
        "fsdp_cpu_ram_efficient_loading": True,
        "fsdp_auto_wrap_policy": "TRANSFORMER_BASED_WRAP",
        "fsdp_transformer_layer_cls_to_wrap": layer_to_wrap,
        "fsdp_state_dict_type": "FULL_STATE_DICT",
        "fsdp_sharding_strategy": "FULL_SHARD",
    }

@dataclass
class BaseConfig:
    base_model: str
    lora_r: int
    lora_alpha: int
    lora_dropout: float
    gradient_accumulation_steps: int
    micro_batch_size: int
    num_epochs: int
    learning_rate: float
    warmup_steps: int
    evals_per_epoch: int
    weight_decay: float
    model_type: str
    fsdp_config: FsdpConf | None

@dataclass
class Config(BaseConfig):
    wandb_project: str | None = None

    def generate(self, dataset: Dataset[Convos]):
        eval = dataset.eval if dataset.eval else []
        return FullConfig(
            base_model=self.base_model,
            datasets=[convert_dataset(ds) for ds in dataset.train],
            test_datasets=[convert_dataset(ds) for ds in eval],
            model_type=self.model_type,
            lora_r=self.lora_r,
            lora_alpha=self.lora_alpha,
            lora_dropout=self.lora_dropout,
            gradient_accumulation_steps=self.gradient_accumulation_steps,
            micro_batch_size=self.micro_batch_size,
            num_epochs=self.num_epochs,
            learning_rate=self.learning_rate,
            warmup_steps=self.warmup_steps,
            evals_per_epoch=self.evals_per_epoch,
            weight_decay=self.weight_decay,
            fsdp_config=self.fsdp_config,
            wandb_project=self.wandb_project,
            fsdp=None if not self.fsdp_config else [
                "full_shard",
                "auto_wrap",
            ],
        )

def convert_dataset(dataset: HubMessageConvos | JsonlConvos | HubInstructConvos):
    if isinstance(dataset, HubMessageConvos):
        return AxolotlHubConvos(
            path=dataset.name,
            field_messages=dataset.messages_field,
            message_property_mappings={
                "role": dataset.role_field,
                "content": dataset.content_field,
            },
            roles={
                "user": [dataset.user_role, "user"],
                "assistant": [dataset.assistant_role, "assistant"],
                "system": [dataset.system_role, "system"],
            },
        )
    elif isinstance(dataset, HubInstructConvos):
        return AxolotlHubInstructions(
            path=dataset.name,
            type={
                "field_input": dataset.input_field,
                "field_instruction": dataset.instruction_field,
                "field_output": dataset.output_field,
            },
        )
    else:
        return AxolotlJsonlConvos(path=dataset.path)

@dataclass
class FullConfig(BaseConfig):
    datasets: Sequence[AxolotlJsonlConvos | AxolotlHubConvos | AxolotlHubInstructions]
    test_datasets: Sequence[AxolotlJsonlConvos | AxolotlHubConvos | AxolotlHubInstructions]
    model_type: str
    wandb_project: str | None
    fsdp: Sequence[str] | None
    output_dir: str = "/workspace/training-output"

    tokenizer_type: str = "AutoTokenizer"

    load_in_8bit: bool = False
    load_in_4bit: bool = False
    strict: bool = False

    dataset_prepared_path: str = "last_run_prepared"

    adapter: str = "lora"

    sequence_len: int = 8192
    sample_packing: bool = True
    eval_sample_packing: bool = True
    pad_to_sequence_len: bool = True

    lora_target_linear: bool = True

    optimizer: str = "adamw_torch_fused"
    lr_scheduler: str = "linear"

    train_on_inputs: bool = False
    group_by_length: bool = False
    bf16: str = "auto"
    tf32: bool = False

    gradient_checkpointing: bool = True
    gradient_checkpointing_kwargs: dict[str, bool] = field(default_factory=lambda: {
        "use_reentrant": True
    })
    logging_steps: int = 1
    flash_attention: bool = True

    saves_per_epoch: int = 1
    special_tokens: dict[str, str] = field(default_factory=lambda: {
        "pad_token": "<|end_of_text|>"
    })

    def save(self, output_dir: str):
        clone = replace(self)
        for convos in clone.datasets:
            rewrite_ds_paths(convos)
        for convos in clone.test_datasets:
            rewrite_ds_paths(convos)
        file_contents = yaml.dump(asdict(clone), default_flow_style=False)
        path = os.path.join(output_dir, "config.yaml")
        with open(path, "w") as f:
            f.write(file_contents)

def rewrite_ds_paths(convos: AxolotlJsonlConvos | AxolotlHubConvos | AxolotlHubInstructions):
    if isinstance(convos, AxolotlHubConvos):
        return
    if isinstance(convos, AxolotlHubInstructions):
        return
    convos.path = "./" + convos.path

@dataclass
class CloudConfig:
    provider: Literal["modal"]
    gpu: Literal["l40s"] | Literal["a100"] | Literal["h100"]
    gpu_count: int
    timeout: int
    env: Sequence[str] = field(default_factory=lambda: [
        "WANDB_API_KEY",
        "HF_TOKEN",
    ])

    volumes: Sequence[dict] = field(default_factory=lambda: [
        {"name": "axolotl-cache", "mount": "/workspace/cache"},
        {"name": "axolotl-data", "mount": "/workspace/data"},
        {"name": "axolotl-out", "mount": "/workspace/training-output"},
    ])

    def save(self, output_dir: str):
        path = os.path.join(output_dir, "cloud_config.yaml")
        with open(path, "w") as f:
            f.write(yaml.dump(asdict(self), default_flow_style=False))

@dataclass
class TrainingConfig:
    axolotl_config: FullConfig
    cloud_config: CloudConfig | None

    def save(self, output_dir: str):
        self.axolotl_config.save(output_dir)
        if self.cloud_config:
            self.cloud_config.save(output_dir)



from dataclasses import dataclass
import os
from typing import Literal, Sequence, cast, Any
from .datasets import Dataset, Convos, HubMessageConvos, get_split_name, HubInstructConvos
from .lora import LoraSettings
from together import Together
from datasets import load_dataset
import json

def llama_3_1_8b_together(
    dataset: Dataset[Convos],
    settings: LoraSettings,
    api_key: str,
    output_dir: str,
):
    """Given a dataset and some settings, returns a Together config to train a
    Llama 3.1 8B Instruct LoRA"""
    return TogetherConfig(
        base_model="meta-llama/Meta-Llama-3.1-8B-Instruct-Reference",
        api_key=api_key,
        lora_settings=settings,
        dataset=dataset,
        output_dir=output_dir,
    )

def llama_3_1_70b_together(
    dataset: Dataset[Convos],
    settings: LoraSettings,
    api_key: str,
    output_dir: str,
):
    """Given a dataset and some settings, returns a Together config to train a
    Llama 3.1 8B Instruct LoRA"""
    return TogetherConfig(
        base_model="meta-llama/Meta-Llama-3.1-70B-Instruct-Reference",
        api_key=api_key,
        lora_settings=settings,
        dataset=dataset,
        output_dir=output_dir,
    )

@dataclass
class TogetherUploadData:
    train_file_id: str
    eval_file_id: str | None

@dataclass
class TogetherConfig:
    base_model: str
    api_key: str
    lora_settings: LoraSettings
    dataset: Dataset[Convos]
    output_dir: str
    warmup_ratio: float = 0.0
    batch_size: int | Literal["max"] = "max"
    n_checkpoints: int = 1

    def upload_files(self):
        client = Together(api_key=self.api_key)

        train_path = os.path.join(self.output_dir, "together-train.jsonl")
        with open(train_path, "w") as f:
            for line in convo_lines(self.output_dir, self.dataset.train):
                f.write(line)
                f.write("\n")
        train_file_id = client.files.upload(file=train_path).id
        eval_file_id = None

        if self.dataset.eval is not None:
            eval_path = os.path.join(self.output_dir, "together-eval.jsonl")
            with open(eval_path, "w") as f:
                for line in convo_lines(self.output_dir, self.dataset.eval):
                    f.write(line)
                    f.write("\n")
            eval_file_id = client.files.upload(file=eval_path).id

        return TogetherUploadData(train_file_id, eval_file_id)

    def finetune(self, upload_data: TogetherUploadData):
        client = Together(api_key=self.api_key)
        client.fine_tuning.create(
            training_file=upload_data.train_file_id,
            validation_file=upload_data.eval_file_id or "",
            model=self.base_model,
            batch_size=self.batch_size,
            lora=True,
            lora_r=self.lora_settings.rank,
            n_epochs=self.lora_settings.num_epochs,
            n_evals=self.lora_settings.evals_per_epoch * self.lora_settings.num_epochs,
            lora_alpha=self.lora_settings.alpha,
            lora_dropout=self.lora_settings.dropout,
            weight_decay=self.lora_settings.weight_decay,
            learning_rate=self.lora_settings.learning_rate,
            wandb_project_name=self.lora_settings.wandb_project,
            wandb_api_key=self.lora_settings.wandb_api_key,
        )

def role(convo: HubMessageConvos, role_value: str):
    if role_value == convo.assistant_role:
        return "assistant"
    if role_value == convo.user_role:
        return "user"
    if role_value == convo.system_role:
        return "system"
    raise Exception(f"Unknown role {role_value}")

def convo_lines(output_dir: str, convos: Sequence[Convos]):
    """Given a sequence of convos, yields the paths for each one"""
    for convo in convos:
        if isinstance(convo, HubMessageConvos):
            for split in convo.splits:
                dataset = load_dataset(
                    convo.name,
                    split=get_split_name(split),
                )
                for row in dataset:
                    input_messages = json.loads(
                        cast(dict[Any, Any], row)[convo.messages_field]
                    )
                    converted = []
                    for message in input_messages:
                        converted.append({
                            "role": role(convo, message[convo.role_field]),
                            "content": message[convo.content_field],
                        })
                    yield json.dumps({
                        "messages": converted,
                    })

        elif isinstance(convo, HubInstructConvos):
            for split in convo.splits:
                dataset = load_dataset(
                    convo.name,
                    split=get_split_name(split),
                )
                for row in dataset:
                    converted = []
                    casted = cast(dict[str, str], row)
                    converted.append({
                        "role": "user",
                        "content": f"{casted[convo.instruction_field]}\n{casted[convo.input_field]}",
                    })
                    converted.append({
                        "role": "assistant",
                        "content": casted[convo.output_field]
                    })
                    yield json.dumps({
                        "messages": converted,
                    })
        else:
            with open(os.path.join(output_dir, convo.path)) as f:
                for line in f:
                    yield line.strip()



import asyncio
import aiohttp
import json
from typing import Any, Protocol, Sequence, TypedDict, Literal
from dataclasses import dataclass

class Message(TypedDict):
    role: Literal["user"] | Literal["assistant"]
    content: str

class ChatClient(Protocol):
    async def chat(
        self,
        prompt: str
    ) -> Sequence[Message]: ...

@dataclass
class OpenAiCompatClient(ChatClient):
    """An OpenAI-compatible API client"""
    api_key: str
    base_url: str
    model: str
    temperature: float = 1.0
    retries: int = 3
    max_tokens: int | None = None

    async def chat(
        self,
        prompt: str
    ) -> Sequence[Message]:
        url = f"{self.base_url}/chat/completions"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}"
        }
        content = ""

        async for line in make_request(
            url=url,
            headers=headers,
            payload={
                "model": self.model,
                "messages": [{"role": "user", "content": prompt}],
                "stream": True,
                "temperature": self.temperature,
                "max_tokens": self.max_tokens,
            },
            retries=self.retries,
        ):
            json_data = json.loads(line)
            if (
                "choices" in json_data
                and json_data["choices"]
                and "delta" in json_data["choices"][0]
                and "content" in json_data["choices"][0]["delta"]
            ):
                content += json_data["choices"][0]["delta"]["content"]
            elif "error" in json_data:
                raise Exception(json_data["error"])

        return [
            {
                "role": "user",
                "content": prompt,
            },
            {
                "role": "assistant",
                "content": content
            },
        ]

@dataclass
class AnthropicCompatClient(ChatClient):
    """An Anthropic-compatible API client. If thinking_budget is set, streams
    down thinking tokens and outputs them inside <think> tags in the assistant
    response object"""
    api_key: str
    model: str
    max_tokens: int
    thinking_budget: None | int = None
    base_url: str = "https://api.anthropic.com/v1"
    temperature: float = 1.0
    retries: int = 3

    async def chat(
        self,
        prompt: str
    ) -> Sequence[Message]:
        url = f"{self.base_url}/messages"
        headers = {
            "Content-Type": "application/json",
            "x-api-key": self.api_key,
            "anthropic-version": "2023-06-01",
        }
        content = ""
        thoughts = None
        async for line in make_request(
            url=url,
            headers=headers,
            payload={
                "model": self.model,
                "messages": [{"role": "user", "content": prompt}],
                "stream": True,
                "temperature": self.temperature,
                "max_tokens": self.max_tokens,
                "thinking": None if self.thinking_budget is None else {
                    "type": "enabled",
                    "budget_tokens": self.thinking_budget,
                }
            },
            retries=self.retries,
        ):
            json_data = json.loads(line)
            if "type" in json_data:
                if json_data["type"] == "error":
                    raise Exception(json_data["error"])
                elif json_data["type"] == "content_block_delta":
                    if json_data["delta"]["type"] == "text_delta":
                        content += json_data["delta"]["text"]
                    if json_data["delta"]["type"] == "thinking_delta":
                        if thoughts is None:
                            thoughts = json_data["delta"]["thinking"]
                        else:
                            thoughts = thoughts + json_data["delta"]["thinking"]

        return [
            {
                "role": "user",
                "content": prompt,
            },
            {
                "role": "assistant",
                "content": f"<think>{thoughts}</think>\n{content}" if thoughts else content
            },
        ]

async def make_request(
    url: str,
    headers: dict[str, str],
    payload: Any,
    retries: int,
):
    for attempt in range(retries):
        try:
            got_bytes = False
            async with aiohttp.ClientSession() as session:
                async with session.post(url, headers=headers, json=payload) as response:
                    async for linebytes in response.content:
                        line = linebytes.decode('utf-8').strip()
                        if line:
                            # SSE format starts with "data: "
                            if line.startswith('data: '):
                                if line == 'data: [DONE]':
                                    break
                                got_bytes = True
                                yield line[6:]
                            elif line.startswith('{'):
                                got_bytes = True
                                yield line
            if got_bytes:
                return
        except Exception as e:
            if attempt == retries - 1:
                print(f"Failed after {retries} attempts: {e}")
                raise
            await asyncio.sleep(1 * (attempt + 1))



from datasets import load_dataset, Dataset as HfDataset
from dataclasses import dataclass
from typing import Sequence, TypeVar, Generic, Iterator, cast, Dict
from collections.abc import Callable
import json

T = TypeVar("T")
@dataclass
class Dataset(Generic[T]):
    train: Sequence[T]
    eval: Sequence[T] | None = None

    def merge(self, dataset: 'Dataset[T]'):
        """Merges this dataset with a given dataset, returning a new dataset
        that's the merge of both"""
        eval = []
        if self.eval is not None:
            eval = [ e for e in self.eval ]
        if dataset.eval is not None:
            eval = eval + [ e for e in dataset.eval ]

        return Dataset(
            train=[
                t for t in self.train
            ] + [
                t for t in dataset.train
            ],
            eval=eval if len(eval) > 0 else None,
        )

@dataclass
class HubSplit:
    name: str
    max_rows: int | None = None

@dataclass
class HubSubset:
    name: str
    split: str | HubSplit

@dataclass
class Prompts:
    output_path: str
    count: Callable[[], int]
    items: Callable[[], Iterator[str]]

def hub_subsets(name: str, subsets: Sequence[HubSubset], text_field: str):
    """Given a Hugging Face hub dataset name and a sequence of subsets, returns
    a set of prompts from the dataset subsets"""
    datasets = [
        cast(HfDataset, load_dataset(
            name,
            split=get_split_name(subset.split),
            data_dir=subset.name
        )) for subset in subsets
    ]
    def count():
        rows = 0
        for i in range(len(datasets)):
            curr_split = subsets[i].split
            if isinstance(curr_split, HubSplit):
                if curr_split.max_rows:
                    rows += curr_split.max_rows
                    continue
            rows += len(datasets[i])
        return rows

    def items():
        for i in range(len(datasets)):
            dataset = datasets[i]
            subset = subsets[i]
            count = 0
            max_rows = subset.split.max_rows if isinstance(subset.split, HubSplit) else None
            for example in dataset:
                row = cast(Dict[str, str], example)
                yield row[text_field]
                count += 1
                if (max_rows is not None) and count >= max_rows:
                    break

    output_path = f"hub/{name}"
    for subset in subsets:
        split_name = get_split_name(subset.split)
        output_path = f"{output_path}-{subset.name}-{split_name}"

    return Prompts(
        output_path=f"{output_path}.jsonl",
        count=count,
        items=items,
    )

def hub_prompts(name: str, split: str | HubSplit, text_field: str):
    """Given a Hugging Face hub dataset name and a split, returns a set of
    prompts from the split"""
    split_name = get_split_name(split)
    ds = cast(HfDataset, load_dataset(name, split=split_name))
    max_rows = split.max_rows if isinstance(split, HubSplit) else None

    def count():
        return max_rows or len(ds)

    def items():
        count = 0
        for example in ds:
            row = cast(Dict[str, str], example)
            yield row[text_field]
            count += 1
            if (max_rows is not None) and count >= max_rows:
                break

    return Prompts(
        output_path=hub_split_path(name=name, split=split_name),
        count=count,
        items=items,
    )

def jsonl_prompts(path: str, name: str, text_field: str):
    """Given a path to a JSONL file of prompts, returns a set of prompts
    extracted from the JSONL file"""
    def count():
        with open(path, 'r') as f:
            total_lines = sum(1 for _ in f)
            return total_lines

    def items():
        with open(path, 'r') as f:
            for line in f:
                data = json.loads(line.strip())
                yield data[text_field]

    return Prompts(
        output_path=f"jsonl/{name}.jsonl",
        count=count,
        items=items,
    )

@dataclass
class JsonlConvos:
    """Trainable conversations in JSONL files"""
    path: str

@dataclass
class HubMessageConvos:
    """Trainable conversations in Hugging Face hub datasets in conversational
    formats"""
    name: str
    splits: Sequence[str | HubSplit]
    messages_field: str = "messages"
    role_field: str = "role"
    content_field: str = "content"
    assistant_role: str = "assistant"
    user_role: str = "user"
    system_role: str = "system"

@dataclass
class HubInstructConvos:
    """Trainable conversations in Hugging Face hub datasets in instructional
    formats"""
    name: str
    splits: Sequence[str | HubSplit]
    instruction_field: str = "instruction"
    input_field: str = "input"
    output_field: str = "output"

Convos = JsonlConvos | HubMessageConvos | HubInstructConvos

def hub_split_path(name: str, split: str):
    return f"hub/{name}-{split}.jsonl"

def get_split_name(split: str | HubSplit):
    if isinstance(split, HubSplit):
        return split.name
    return split
